#include "skein/orbit.h"

#include "astro/stateVectorIndices.hpp"
#include "astro/orbitalElementConversions.hpp"

#include <glm/gtc/matrix_transform.hpp>

Orbit::Orbit()
{
    _keplerianElements.resize(6);
}

float Orbit::getSemiMajorAxis() const
{
    return _keplerianElements[astro::semiMajorAxisIndex];
}
void Orbit::setSemiMajorAxis(float semiMajorAxis)
{
    _keplerianElements[astro::semiMajorAxisIndex] = semiMajorAxis;
}

float Orbit::getEccentricity() const
{
    return _keplerianElements[astro::eccentricityIndex];
}
void Orbit::setEccentricity(float eccentricity)
{
    _keplerianElements[astro::eccentricityIndex] = eccentricity;
}

float Orbit::getInclination() const
{
    return _keplerianElements[astro::inclinationIndex];
}
void Orbit::setInclination(float inclination)
{
    _keplerianElements[astro::inclinationIndex] = inclination;
}

float Orbit::getArgumentOfPeriapsis() const
{
    return _keplerianElements[astro::argumentOfPeriapsisIndex];
}
void Orbit::setArgumentOfPeriapsis(float argumentOfPeriapsis)
{
    _keplerianElements[astro::argumentOfPeriapsisIndex] = argumentOfPeriapsis;
}

float Orbit::getLongitudeOfAscendingNode() const
{
    return _keplerianElements[astro::longitudeOfAscendingNodeIndex];
}
void Orbit::setLongitudeOfAscendingNode(float longitudeOfAscendingNode)
{
    _keplerianElements[astro::longitudeOfAscendingNodeIndex] = longitudeOfAscendingNode;
}

glm::mat4 Orbit::getLookAlongMatrix(const float meanAnomaly) const
{
    glm::vec3 position = getPosition(meanAnomaly);

    // get the tangent of the orbital ellipse
    glm::vec3 tan = getTangent(meanAnomaly);
    // we want to point along the orbit
    glm::vec3 target = position + tan;
    // TODO: this is not 'up' with respect to the orbited body!
    // 'up' is just the normalized position vector because the orbit is centred at the origin
    glm::vec3 up = glm::normalize(position);
    // easy peasy lookAt matrix
    glm::mat4 look = glm::lookAt(position, target, up);

    // invert the lookat matrix because it's meant for cameras, cameras work backwards and
    // we are not a camera
    glm::mat4 lookAlong = glm::inverse(look);

    return lookAlong;
}

const float Orbit::getEccentricAnomaly(const float meanAnomaly) const
{
    const float eccentricity = _keplerianElements[astro::eccentricityIndex];
    float eccentricAnomaly = astro::convertEllipticalMeanAnomalyToEccentricAnomalyBS(
        eccentricity,
        meanAnomaly,
        100);

    return eccentricAnomaly;
}

// Interpolate a position around the orbit.
// t is in range 0..1 and wraps.
glm::vec3 Orbit::getPosition(const float meanAnomaly) const
{
    Vector6 cartesian = getCartesianCoordinates(meanAnomaly);
    return glm::vec3(
        cartesian[astro::xPositionIndex],
        cartesian[astro::yPositionIndex],
        cartesian[astro::zPositionIndex]);
}

const float Orbit::getTrueAnomaly(const float meanAnomaly) const
{
    const float eccentricAnomaly = getEccentricAnomaly(meanAnomaly);
    const float eccentricity = _keplerianElements[astro::eccentricityIndex];
    return astro::convertEccentricAnomalyToTrueAnomaly(
        eccentricAnomaly,
        eccentricity);
}

const Vector6 Orbit::getCartesianCoordinates(const float meanAnomaly) const
{
    Vector6 kepler(_keplerianElements);
    kepler[astro::trueAnomalyIndex] = getTrueAnomaly(meanAnomaly);
    return astro::convertKeplerianToCartesianElements(kepler, 1.0);
}

glm::vec3 Orbit::getTangent(const float meanAnomaly) const
{
    float epsilon = 0.01;
    glm::vec3 ahead = getPosition(meanAnomaly + epsilon);
    glm::vec3 behind = getPosition(meanAnomaly - epsilon);
    return glm::normalize(ahead - behind);
}

